#include "Scene.h"
#include "Config.h"
#include "Raytrace.h"
#include "Sphere.h"
#include <iostream>
using namespace std;

const vecteur NullVector = { 0.0f,0.0f,0.0f };
const point Origin = { 0.0f,0.0f,0.0f };
const SimpleString emptyString("");
const SimpleString typeString("Type");
const SimpleString bumplevelString("BumpLevel");
const SimpleString diffuseString ("Diffuse");
const SimpleString diffuse2String ("Diffuse2");

void GetMaterial(const Config &sceneFile, material &currentMat)
{
    SimpleString materialType = sceneFile.GetByNameAsString(typeString, emptyString);

    if (materialType.compare("turbulence") == 0)
    {
        currentMat.type = material::turbulence;
    }
    else if (materialType.compare("marble") == 0)
    {
        currentMat.type = material::marble;
    }
    else if (materialType.compare("noise") == 0)
    {
        currentMat.type = material::noise;
    }
    else
    { 
        // default
        currentMat.type = material::gouraud;
    }
    
    currentMat.bump = float(sceneFile.GetByNameAsFloat(bumplevelString, 0.0f));

    materialChannel & redChannel = currentMat.getChannel(color::OFFSET_RED);
    materialChannel & greenChannel = currentMat.getChannel(color::OFFSET_GREEN);
    materialChannel & blueChannel = currentMat.getChannel(color::OFFSET_BLUE);

    // diffuse color
    {
        float fScalar;
        fScalar =  float(sceneFile.GetByNameAsFloat(diffuseString, 0.0f)); 
        vecteur vColor = {fScalar, fScalar, fScalar};
        vColor = sceneFile.GetByNameAsVector(diffuseString, vColor);
        redChannel.diffuse = vColor.x;
        greenChannel.diffuse = vColor.y;
        blueChannel.diffuse = vColor.z;
    }

    // diffuse2 color
    {
        float fScalar;
        fScalar =  float(sceneFile.GetByNameAsFloat(diffuse2String, 0.0f)); 
        vecteur vColor = {fScalar, fScalar, fScalar};
        vColor = sceneFile.GetByNameAsVector(diffuse2String, vColor);
        redChannel.diffuse2 = vColor.x;
        greenChannel.diffuse2 = vColor.y;
        blueChannel.diffuse2 = vColor.z;
    }

    // Reflexion color
    {
        float fScalar;
        fScalar =  float(sceneFile.GetByNameAsFloat("Reflection", 0.0f)); 
        vecteur vColor = {fScalar, fScalar, fScalar};
        vColor = sceneFile.GetByNameAsVector("Reflection", vColor);
        redChannel.reflection = vColor.x;
        greenChannel.reflection = vColor.y;
        blueChannel.reflection = vColor.z;
    }

    // Refraction color
    {
        float fScalar;
        fScalar =  float(sceneFile.GetByNameAsFloat("Refraction", 0.0f)); 
        vecteur vColor = {fScalar, fScalar, fScalar};
        vColor = sceneFile.GetByNameAsVector("Refraction", vColor);
        redChannel.refraction = vColor.x;
        greenChannel.refraction = vColor.y;
        blueChannel.refraction = vColor.z;
    }

    // Density color
    {
        float fScalar;
        fScalar =  float(sceneFile.GetByNameAsFloat("Density", 0.0f)); 
        vecteur vColor = {fScalar, fScalar, fScalar};
        vColor = sceneFile.GetByNameAsVector("Density", vColor);
        redChannel.density = vColor.x;
        greenChannel.density = vColor.y;
        blueChannel.density = vColor.z;
    }

    // Absorption color
    {
        float fScalar;
        fScalar =  float(sceneFile.GetByNameAsFloat("Absorption", 0.0f)); 
        vecteur vColor = {fScalar, fScalar, fScalar};
        vColor = sceneFile.GetByNameAsVector("Absorption", vColor);
        redChannel.absorption = vColor.x;
        greenChannel.absorption= vColor.y;
        blueChannel.absorption = vColor.z;
    }

    // Specular color
    {
        float fScalar;
        fScalar =  float(sceneFile.GetByNameAsFloat("Specular", 0.0f)); 
        vecteur vColor = {fScalar, fScalar, fScalar};
        vColor = sceneFile.GetByNameAsVector("Specular", vColor);
        redChannel.specular = vColor.x;
        greenChannel.specular = vColor.y;
        blueChannel.specular = vColor.z;
    }

    // Specular Power color
    {
        float fScalar;
        fScalar =  float(sceneFile.GetByNameAsFloat("Power", 0.0f)); 
        vecteur vColor = {fScalar, fScalar, fScalar};
        vColor = sceneFile.GetByNameAsVector("Power", vColor);
        redChannel.power = vColor.x;
        greenChannel.power = vColor.y;
        blueChannel.power = vColor.z;
    }

}

void GetSphere(const Config &sceneFile, sphere &currentSph)
{
    currentSph.pos = sceneFile.GetByNameAsPoint("Center", Origin); 

    currentSph.size =  float(sceneFile.GetByNameAsFloat("Size", 0.0f)); 

    currentSph.materialId = sceneFile.GetByNameAsInteger("Material.Id", 0); 
}

void GetBlob(const Config &sceneFile, blob &currentBlb)
{
    currentBlb.centerList.resize(3);
    currentBlb.centerList[0] = sceneFile.GetByNameAsPoint("Center0", Origin); 
    currentBlb.centerList[1] = sceneFile.GetByNameAsPoint("Center1", Origin); 
    currentBlb.centerList[2] = sceneFile.GetByNameAsPoint("Center2", Origin); 

    currentBlb.size =  float(sceneFile.GetByNameAsFloat("Size", 0.0f)); 

    currentBlb.materialId = sceneFile.GetByNameAsInteger("Material.Id", 0); 
}

void GetLight(const Config &sceneFile, light &currentLight)
{
    currentLight.pos = sceneFile.GetByNameAsPoint("Position", Origin); 

    currentLight.intensity.blue =  float(sceneFile.GetByNameAsFloat("Intensity.Blue", 1.0f)); 
    currentLight.intensity.red  =  float(sceneFile.GetByNameAsFloat("Intensity.Red", 1.0f)); 
    currentLight.intensity.green = float(sceneFile.GetByNameAsFloat("Intensity.Green", 1.0f)); 
}

void GetPhotonMapSettings (const Config &sceneFile, photonMapSettings & settings)
{
    SimpleString sFilterType = sceneFile.GetByNameAsString("FilterType", emptyString);
    if (sFilterType.compare("gaussian") == 0)
    {
        settings.eFilterType = photonMapSettings::gaussianEstimator;  
    }
    else if (sFilterType.compare("conic") == 0)
    {
        settings.eFilterType = photonMapSettings::conicEstimator;  
    }
    else // default
    {
        settings.eFilterType = photonMapSettings::radianceEstimator;  
    }

    settings.maxDistanceSquare = sceneFile.GetByNameAsFloat("MaxDistanceSquare", 0.0f);
    settings.directLighting = sceneFile.GetByNameAsBoolean("DirectLighting", false);
    settings.maxNearestPhotons = sceneFile.GetByNameAsInteger("MaxNearestPhotons", 0);
    settings.size = sceneFile.GetByNameAsInteger("Size", 0);
}

bool init(char* inputName, scene &myScene)
{
	int nbMats, nbSpheres, nbBlobs, nbLights, nbPhotonMaps;
	int i;
	Config sceneFile(inputName);
    if (sceneFile.SetSection("Scene") == -1)
    {
        cout << "Fichier scene mal form" << endl;
		return false;
    }
    myScene.sizex = sceneFile.GetByNameAsInteger("Image.Width", 640);
    myScene.sizey = sceneFile.GetByNameAsInteger("Image.Height", 480);

    {
        SimpleString supersampleType = sceneFile.GetByNameAsString("Image.SupersampleType", emptyString);

        if (supersampleType.compare("high") == 0)
        {
            myScene.supersampleType = scene::high;
        }
        else if (supersampleType.compare("medium") == 0)
        {
            myScene.supersampleType = scene::medium;
        }
        else // if (supersampleType.compare("low") == 0)
        {
            myScene.supersampleType = scene::low;
        }
    }

    myScene.complexity = sceneFile.GetByNameAsInteger("Complexity", 1);

    nbMats = sceneFile.GetByNameAsInteger("NumberOfMaterials", 0);
    nbSpheres = sceneFile.GetByNameAsInteger("NumberOfSpheres", 0);
    nbBlobs = sceneFile.GetByNameAsInteger("NumberOfBlobs", 0);
    nbLights = sceneFile.GetByNameAsInteger("NumberOfLights", 0);
    nbPhotonMaps = sceneFile.GetByNameAsInteger("NumberOfPhotonMaps", 0);

    myScene.cm.name[cubemap::up] = sceneFile.GetByNameAsString("Cubemap.Up", emptyString);
    myScene.cm.name[cubemap::down] = sceneFile.GetByNameAsString("Cubemap.Down", emptyString);
    myScene.cm.name[cubemap::right] = sceneFile.GetByNameAsString("Cubemap.Right", emptyString);
    myScene.cm.name[cubemap::left] = sceneFile.GetByNameAsString("Cubemap.Left", emptyString);
    myScene.cm.name[cubemap::forward] = sceneFile.GetByNameAsString("Cubemap.Forward", emptyString);
    myScene.cm.name[cubemap::backward] = sceneFile.GetByNameAsString("Cubemap.Backward", emptyString);
    myScene.cm.bExposed = sceneFile.GetByNameAsBoolean("Cubemap.Exposed", false);
    myScene.cm.bsRGB = sceneFile.GetByNameAsBoolean("Cubemap.sRGB", false);
    myScene.cm.exposure = float(sceneFile.GetByNameAsFloat("Cubemap.Exposure", 1.0f));

    bool bHDR = sceneFile.GetByNameAsBoolean("Cubemap.HDR", false);

    if (bHDR)
    {
        if (!myScene.cm.InitHDR())
        {
            cout << "Pas de fichier skybox HDR trouv" << endl;
        }
    }
    else if (!myScene.cm.Init())
    {
        cout << "Pas de fichier skybox trouv" << endl;
    }

    {
        SimpleString perspectiveType = sceneFile.GetByNameAsString("Perspective.Type", emptyString);
        if (perspectiveType.compare("conic") == 0)
        {
            myScene.persp.type = perspective::conic;
            myScene.persp.FOV = float(sceneFile.GetByNameAsFloat("Perspective.FOV", 45.0f));
            myScene.persp.clearPoint = float(sceneFile.GetByNameAsFloat("Perspective.ClearPoint", 1.0f));
            myScene.persp.dispersion = float(sceneFile.GetByNameAsFloat("Perspective.Dispersion", 0.0f));
        }
        else
        { 
            // default
            myScene.persp.type = perspective::orthogonal;
        }
    }

	myScene.matTab.resize(nbMats);
	myScene.sphTab.resize(nbSpheres);
	myScene.blbTab.resize(nbBlobs);
	myScene.lgtTab.resize(nbLights);

	for (i=0; i<nbMats; ++i)
    {   
        material &currentMat = myScene.matTab[i];
        SimpleString sectionName("Material");
        sectionName.append((unsigned long) i);
        if (sceneFile.SetSection( sectionName ) == -1)
        {
            cout << "Fichier scene mal form" << endl;
		    return false;
        }
        GetMaterial(sceneFile, currentMat);
    }
	for (i=0; i<nbSpheres; ++i)
    {   
        sphere &currentSphere = myScene.sphTab[i];
        SimpleString sectionName("Sphere");
        sectionName.append((unsigned long) i);
        if (sceneFile.SetSection( sectionName ) == -1)
        {
            cout << "Fichier scene mal form" << endl;
		    return false;
        }
        GetSphere(sceneFile, currentSphere);
        if (currentSphere.materialId >= nbMats)
        {
            cout << "Fichier scene mal form" << endl;
		    return false;
        }

    }
	for (i=0; i<nbBlobs; ++i)
    {   
        blob &currentBlob = myScene.blbTab[i];
        SimpleString sectionName("Blob");
        sectionName.append((unsigned long) i);
        if (sceneFile.SetSection( sectionName ) == -1)
        {
            cout << "Fichier scene mal form" << endl;
		    return false;
        }
        GetBlob(sceneFile, currentBlob);
        if (currentBlob.materialId >= nbMats)
        {
            cout << "Fichier scene mal form" << endl;
		    return false;
        }
    }
	for (i=0; i<nbLights; ++i)
    {   
        light &currentLight = myScene.lgtTab[i];
        SimpleString sectionName("Light");
        sectionName.append((unsigned long) i);
        if (sceneFile.SetSection( sectionName ) == -1)
        {
            cout << "Fichier scene mal form" << endl;
		    return false;
        }
        GetLight(sceneFile, currentLight);
        
    }
    if (nbPhotonMaps)
    {
        SimpleString sectionName("Photonmap0");
        if (sceneFile.SetSection( sectionName ) == -1)
        {
            cout << "Fichier scene mal form" << endl;
		    return false;
        }
        GetPhotonMapSettings(sceneFile, myScene.phMapProp);
    }

	return true;
}

