#include "ray.h"
#include "scene.h"
#include "Raytrace.h"
#include "Sphere.h"
#include "blob.h"
#include <cmath>

bool findIntersection(/*IN*/const ray viewRay, 
                      /*IN*/const scene & myScene, 
                      /*OUT*/point &ptHitPoint, 
                      /*OUT*/vecteur &vNormal, 
                      /*OUT*/material &currentMat, 
                      /*OUT*/float &distance)
{
    int currentBlob=-1;
    int currentSphere=-1;
    distance = 20000.0f;
    for (unsigned int i = 0; i < myScene.blbTab.size() ; ++i)
    {
        if (isBlobIntersected(viewRay, myScene.blbTab[i], distance)) {
            currentBlob = i;
        }
    }
    for (unsigned int i = 0; i < myScene.sphTab.size() ; ++i)
    {
        if (hitSphere(viewRay, myScene.sphTab[i], distance))
        {
            currentSphere = i;
            currentBlob = -1;
        }
    }
    if (currentBlob != -1)
    {
        ptHitPoint  = viewRay.start + distance * viewRay.dir;
        blobInterpolation(ptHitPoint, myScene.blbTab[currentBlob], &myScene.matTab[0], vNormal, currentMat);
        float temp = vNormal * vNormal;
        if (temp == 0.0f)
            return false;
        vNormal = invsqrtf(temp) * vNormal;
        currentMat = myScene.matTab[myScene.blbTab[currentBlob].materialId];
        return true;
    }
    else if (currentSphere != -1)
    {
        ptHitPoint  = viewRay.start + distance * viewRay.dir;
		vNormal = ptHitPoint - myScene.sphTab[currentSphere].pos;
		float temp = vNormal * vNormal;
		if (temp == 0.0f)
			return false;
		temp = invsqrtf(temp);
		vNormal = temp * vNormal;
		currentMat = myScene.matTab[myScene.sphTab[currentSphere].materialId];
        return true;
    }
    
    return false;
}

bool LookForDiffuse(/*IN*/const ray &viewRay,
                   /*IN*/const point &ptHitPoint, 
                   /*IN*/const vecteur &vNormal, 
                   /*IN*/const materialChannel &currentMatChannel, 
                   /*IN*/ bool bInside,
                   /*IN-OUT*/ context &myContext,
                   /*OUT*/ ray &outRay)
{    
    float fViewProjection = viewRay.dir * vNormal;
    float fReflectance, fTransmittance;
    float fCosThetaI, fSinThetaI, fCosThetaT, fSinThetaT;

    if(((currentMatChannel.reflection != 0.0f) || (currentMatChannel.refraction != 0.0f) ) && (currentMatChannel.density != 0.0f))
    {
        // material de type verre on va calculer le coefficient de reflexion de fresnel

        float fDensity1 = myContext.fRefractionCoef; 
        float fDensity2;
        if (bInside)
        {
            // on considere que le rayon incident traverse un ether proche du vide (ou de l'air)
            // en thorie il faudrait d'abord calculer pour savoir si l'extrieur
            // se trouve dans un autre objet mais a dpasse le cadre de notre code.
            fDensity2 = context::getDefaultAir().fRefractionCoef;
        }
        else
        {
            fDensity2 = currentMatChannel.density;
        }

        // ici on tient compte du fait que le dplacement de la lumiere
        // est symetrique, de l'observateur a la source de la lumiere ou de la source de lumiere
        // a l'observateur. On fait donc les calculs du coefficient en prenant en compte le rayon provenant 
        // du point d'observation.
        fCosThetaI = fabsf(fViewProjection); 

        if (fCosThetaI >= 0.999f) 
        {
            // cas du rayon incident selon la normale a la surface
            fReflectance = (fDensity1 - fDensity2) / (fDensity1 + fDensity2);
            fReflectance = fReflectance * fReflectance;
            fSinThetaI = 0.0f;
            fSinThetaT = 0.0f;
            fCosThetaT = 1.0f;
        }
        else 
        {
            fSinThetaI = sqrtf(1 - fCosThetaI * fCosThetaI);
            // le signe de Sin ThetaI n'a aucune importance, c'est le meme que celui de Sin ThetaT
            // et ils s'annihilent mutuellement dans le calcul des coefficients de Reflexion
            fSinThetaT = (fDensity1 / fDensity2) * fSinThetaI;
            if (fSinThetaT * fSinThetaT > 0.9999f)
            {
                // On a atteint l'angle critique au dela duquel
                // la surface est entierement reflexive
                fReflectance = 1.0f ;
                fCosThetaT = 0.0f;
            }
            else
            {
                fCosThetaT = sqrtf(1 - fSinThetaT * fSinThetaT);
                // tout d'abord la reflectance de la lumiere polaris dans le 
                // plan orthogonal au plan de reflexion
                float fReflectanceOrtho = (fDensity2 * fCosThetaT - fDensity1 * fCosThetaI ) 
                    / (fDensity2 * fCosThetaT + fDensity1  * fCosThetaI);
                fReflectanceOrtho = fReflectanceOrtho * fReflectanceOrtho;
                // Ensuite la reflectance de la lumiere polaris dans le 
                // plan parallele au plan de reflexion
                float fReflectanceParal = (fDensity1 * fCosThetaT - fDensity2 * fCosThetaI )
                    / (fDensity1 * fCosThetaT + fDensity2 * fCosThetaI);
                fReflectanceParal = fReflectanceParal * fReflectanceParal;

                // le coefficient de reflectance global est gal  la moyenne
                // des deux pour une lumire de base non polarise
                fReflectance =  0.5f * (fReflectanceOrtho + fReflectanceParal);
            }
        }
    }
    else
    {
        // reflexion de type "metal", la reflectance est la meme dans toutes les directions
        // Les mtaux sont conducteurs et induisent un changement dans la polarisation de la lumire
        // que l'on ignore royalement.
        fReflectance = 1.0f;
        fCosThetaI = 1.0f;
        fCosThetaT = 1.0f;
    }

    fTransmittance = currentMatChannel.refraction * (1.0f - fReflectance);
    fReflectance = currentMatChannel.reflection * fReflectance;

    float fTotalWeight = fReflectance + fTransmittance;

    if (fTotalWeight > 0.0f)
    {
        float fRoulette = (1.0f / RAND_MAX) * rand();
    
        if (fRoulette <= fReflectance)
        {
            // Le rayon viewRay est rflechi par la normale.
            // Pour calculer le rayon reflchi on fait la projection
            // du vecteur direction du rayon vue
            // sur la direction de la normale.
            // Pour avoir le vecteur tangent il faudrait retrancher
            // cette projection au vecteur direction
            // mais ici on veut la reflexion et donc il faut retrancher
            // la projection deux fois au vecteur direction.

            float fReflection = - 2.0f * fViewProjection;

            // on fait partir le nouveau rayon du point d'intersection avec la sphre courante
            // et on le dirige selon le vecteur reflexion que l'on vient de calculer
            outRay.start = ptHitPoint;
		    outRay.dir = viewRay.dir + fReflection * vNormal;

            if (bInside) 
            {   
                myContext.absorption = currentMatChannel.absorption;
            }
            else 
            {
                myContext.absorption = context::getDefaultAir().fRefractionCoef;
            }
            return false;
        }
        else if(fRoulette <= fTotalWeight)
        {
            float fOldRefractionCoef = myContext.fRefractionCoef;

            if (bInside) 
            {
                myContext.fRefractionCoef = context::getDefaultAir().fRefractionCoef;
            }
            else
            {
                myContext.fRefractionCoef = currentMatChannel.density;
            }

            // ici on calcule le rayon transmis avec la formule de Snell-Descartes
            outRay.start = ptHitPoint;

            outRay.dir = viewRay.dir + fCosThetaI * vNormal;
            outRay.dir = (fOldRefractionCoef / myContext.fRefractionCoef) * outRay.dir;
            outRay.dir += (-fCosThetaT) * vNormal;

            if (bInside) 
            {   
                myContext.absorption = 1.0f;
            }
            else 
            {
                myContext.absorption = currentMatChannel.absorption;
            }
            return false;
        }
        else
        {
            return true;
        }
    }
    return true;
}
